#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.button module

This module provides base form button widget.
"""

from zope.interface import implementer_only

from pyams_form.browser.widget import HTMLInputWidget, add_field_class
from pyams_form.interfaces.button import IButton
from pyams_form.interfaces.widget import IButtonWidget, IFieldWidget
from pyams_form.widget import Widget, FieldWidget
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@implementer_only(IButtonWidget)
class ButtonWidget(HTMLInputWidget, Widget):
    """A simple button of a form."""

    klass = 'button-widget'
    css = 'button'

    def update(self):
        # We do not need to use the widget's update method, because it is
        # mostly about ectracting the value, which we do not need to do.
        add_field_class(self)


@adapter_config(required=(IButton, IFormLayer), provides=IFieldWidget)
def ButtonFieldWidget(field, request):  # pylint: disable=invalid-name
    """Simple form button widget factory adapter"""
    button = FieldWidget(field, ButtonWidget(request))
    button.value = field.title
    return button



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.term module

Terms management module.

Note: This module doesn't use snake_case for compatibility purposes with zope.schema package,
which implies many Pylint annotations...
"""

from zope.interface import Interface
from zope.schema.interfaces import IBaseVocabulary, IBool, IChoice, ICollection, IIterableSource
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_form.interfaces import IBoolTerms, ITerms, IVocabularyTerms
from pyams_form.interfaces.form import IContextAware
from pyams_form.interfaces.widget import IWidget
from pyams_form.util import create_css_id, to_unicode
from pyams_layer.interfaces import IFormLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.form import IDataManager


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


class Terms:
    """Base implementation for custom ITerms."""

    terms = None

    def getTerm(self, value):  # pylint: disable=invalid-name
        """Get term matching given value"""
        return self.terms.getTerm(value)

    def getTermByToken(self, token):  # pylint: disable=invalid-name
        """Get term matching given token"""
        return self.terms.getTermByToken(token)

    def getValue(self, token):  # pylint: disable=invalid-name
        """Get value matching given token"""
        return self.getTermByToken(token).value

    def __iter__(self):
        return iter(self.terms)

    def __len__(self):
        return self.terms.__len__()

    def __contains__(self, value):
        return self.terms.__contains__(value)


class SourceTerms(Terms):
    """Base implementation for ITerms using a source instead of a vocabulary."""

    def __init__(self, context, request, form, field, source, widget):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.form = form
        self.field = field
        self.widget = widget
        self.source = source
        self.terms = request.registry.getMultiAdapter((self.source, self.request),
                                                      IVocabularyTerms)

    def getTerm(self, value):
        try:
            return super().getTerm(value)
        except KeyError as err:
            raise LookupError(value) from err

    def getTermByToken(self, token):
        # This is rather expensive
        for value in self.source:
            term = self.getTerm(value)
            if term.token == token:
                return term
        raise LookupError(token)

    def getValue(self, token):
        try:
            return self.terms.getValue(token)
        except KeyError as err:
            raise LookupError(token) from err

    def __iter__(self):
        for value in self.source:
            yield self.terms.getTerm(value)

    def __len__(self):
        return len(self.source)

    def __contains__(self, value):
        return value in self.source


@adapter_config(required=(Interface, IFormLayer, Interface, IChoice, IWidget),
                provides=ITerms)
def ChoiceTerms(context, request, form, field, widget):  # pylint: disable=invalid-name
    """Choice terms adapter"""
    if field.context is None:
        field = field.bind(context)
    terms = field.vocabulary
    return request.registry.queryMultiAdapter((context, request, form, field, terms, widget),
                                              ITerms)


@adapter_config(required=(Interface, IFormLayer, Interface, IChoice, IBaseVocabulary, IWidget),
                provides=ITerms)
class ChoiceTermsVocabulary(Terms):
    """ITerms adapter for zope.schema.IChoice based implementations using
    vocabulary."""

    def __init__(self, context, request, form, field, vocabulary, widget):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.form = form
        self.field = field
        self.widget = widget
        self.terms = vocabulary


class MissingTermsBase:  # pylint: disable=no-member
    """Base class for MissingTermsMixin classes."""

    def _can_query_current_value(self):
        return IContextAware.providedBy(self.widget) and not self.widget.ignore_context

    def _query_current_value(self):
        return self.request.registry.getMultiAdapter((self.widget.context, self.field),
                                                     IDataManager).query()

    @staticmethod
    def _make_token(value):
        """create a unique valid ASCII token"""
        return create_css_id(to_unicode(value))

    def _make_missing_term(self, value):
        """Return a term that should be displayed for the missing token"""
        uvalue = to_unicode(value)
        return SimpleTerm(value, self._make_token(value),
                          title=_('Missing: ${value}', mapping=dict(value=uvalue)))


class MissingTermsMixin(MissingTermsBase):
    """This can be used in case previous values/tokens get missing
    from the vocabulary and you still need to display/keep the values"""

    def getTerm(self, value):  # pylint: disable=invalid-name
        """Get term martching given value"""
        try:
            return super().getTerm(value)
        except LookupError:
            if self._can_query_current_value():
                cur_value = self._query_current_value()
                if cur_value == value:
                    return self._make_missing_term(value)
            raise

    def getTermByToken(self, token):  # pylint: disable=invalid-name
        """Get term matching given token"""
        try:
            return super().getTermByToken(token)
        except LookupError as err:
            if self._can_query_current_value():
                value = self._query_current_value()
                term = self._make_missing_term(value)
                if term.token == token:
                    # check if the given token matches the value, if not
                    # fall back on LookupError, otherwise we might accept
                    # any crap coming from the request
                    return term

            raise LookupError(token) from err


@adapter_config(required=(Interface, IFormLayer, Interface, IChoice, IBaseVocabulary, IWidget),
                provides=ITerms)
class MissingChoiceTermsVocabulary(MissingTermsMixin, ChoiceTermsVocabulary):
    """ITerms adapter for zope.schema.IChoice based implementations using
    vocabulary with missing terms support"""


@adapter_config(required=(Interface, IFormLayer, Interface, IChoice, IIterableSource, IWidget),
                provides=ITerms)
class ChoiceTermsSource(SourceTerms):
    """ITerms adapter for zope.schema.IChoice based implementations using source."""


@adapter_config(required=(Interface, IFormLayer, Interface, IChoice, IIterableSource, IWidget),
                provides=ITerms)
class MissingChoiceTermsSource(MissingTermsMixin, ChoiceTermsSource):
    """ITerms adapter for zope.schema.IChoice based implementations using source
       with missing terms support."""


@adapter_config(required=(Interface, IFormLayer, Interface, IBool, IWidget),
                provides=IBoolTerms)
class BoolTerms(Terms):
    """Default yes and no terms are used by default for IBool fields."""

    true_label = _('yes')
    false_label = _('no')

    def __init__(self, context, request, form, field, widget):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.form = form
        self.field = field
        self.widget = widget
        terms = [SimpleTerm(*args)
                 for args in [(True, 'true', self.true_label),
                              (False, 'false', self.false_label)]]
        self.terms = SimpleVocabulary(terms)


@adapter_config(required=(Interface, IFormLayer, Interface, ICollection, IWidget),
                provides=ITerms)
def CollectionTerms(context, request, form, field, widget):  # pylint: disable=invalid-name
    """Collection terms adapter"""
    terms = field.value_type.bind(context).vocabulary
    return request.registry.queryMultiAdapter((context, request, form, field, terms, widget),
                                              ITerms)


@adapter_config(required=(Interface, IFormLayer, Interface, ICollection, IBaseVocabulary, IWidget),
                provides=ITerms)
class CollectionTermsVocabulary(Terms):
    """ITerms adapter for zope.schema.ICollection based implementations using
    vocabulary."""

    def __init__(self, context, request, form, field, vocabulary, widget):
        # pylint: disable=too-many-arguments
        self.context = context
        self.request = request
        self.form = form
        self.field = field
        self.widget = widget
        self.terms = vocabulary


class MissingCollectionTermsMixin(MissingTermsBase):
    """`MissingTermsMixin` adapted to collections."""

    def getTerm(self, value):  # pylint: disable=invalid-name
        """Get term matching given value"""
        try:
            return super().getTerm(value)
        except LookupError:
            if self._can_query_current_value():
                if value in self._query_current_value():
                    return self._make_missing_term(value)
            raise

    def getTermByToken(self, token):  # pylint: disable=invalid-name
        """Get term matching given token"""
        try:
            return super().getTermByToken(token)
        except LookupError:
            if self._can_query_current_value():
                for value in self._query_current_value():
                    term = self._make_missing_term(value)
                    if term.token == token:
                        # check if the given token matches the value, if not
                        # fall back on LookupError, otherwise we might accept
                        # any crap coming from the request
                        return term
            raise

    def getValue(self, token):  # pylint: disable=invalid-name
        """Get value matching given token"""
        try:
            return super().getValue(token)
        except LookupError:
            if self._can_query_current_value():
                for value in self._query_current_value():
                    term = self._make_missing_term(value)
                    if term.token == token:
                        # check if the given token matches the value, if not
                        # fall back on LookupError, otherwise we might accept
                        # any crap coming from the request
                        return value
            raise


class MissingCollectionTermsVocabulary(MissingCollectionTermsMixin,
                                       CollectionTermsVocabulary):
    """ITerms adapter for zope.schema.ICollection based implementations using
    vocabulary with missing terms support."""


@adapter_config(required=(Interface, IFormLayer, Interface, ICollection, IIterableSource, IWidget),
                provides=ITerms)
class CollectionTermsSource(SourceTerms):
    """ITerms adapter for zope.schema.ICollection based implementations using
    source."""


class MissingCollectionTermsSource(MissingCollectionTermsMixin,
                                   CollectionTermsSource):
    """ITerms adapter for zope.schema.ICollection based implementations using
    source with missing terms support."""



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""
from zope.interface import Attribute, Interface
from zope.schema import Field, Object, TextLine

from pyams_form.interfaces import IField, IManager, ISelectionManager
from pyams_form.interfaces.widget import IFieldWidget, IWidget


__docformat__ = 'restructuredtext'

from pyams_form import _


class ActionExecutionError(Exception):
    """An error that occurs during the execution of an action handler."""

    def __init__(self, error):  # pylint: disable=super-init-not-called
        self.error = error

    def __repr__(self):
        return '<%s wrapping %r>' % (self.__class__.__name__, self.error)


class WidgetActionExecutionError(ActionExecutionError):
    """An action execution error that occurred due to a widget value being
    incorrect."""

    def __init__(self, widget_name, error):
        ActionExecutionError.__init__(self, error)
        self.widget_name = widget_name


class IAction(Interface):
    """Action"""

    __name__ = TextLine(title=_('Name'),
                        description=_('The object name.'),
                        required=False,
                        default=None)

    title = TextLine(title=_('Title'),
                     description=_('The action title.'),
                     required=True)

    def is_executed(self):
        """Determine whether the action has been executed."""


class IActionHandler(Interface):
    """Action handler."""


class IActionEvent(Interface):
    """An event specific for an action."""

    action = Object(title=_('Action'),
                    description=_('The action for which the event is created.'),
                    schema=IAction,
                    required=True)


class IActionErrorEvent(IActionEvent):
    """An action event that is created when an error occurred."""

    error = Field(title=_('Error'),
                  description=_('The error that occurred during the action.'),
                  required=True)


class IActions(IManager):
    """A action manager"""

    executed_actions = Attribute('''An iterable of all executed actions (usually just one).''')

    def update(self):
        """Setup actions."""

    def execute(self):
        """Exceute actions.

        If an action execution error is raised, the system is notified using
        the action occurred error; on the other hand, if successful, the
        action successfull event is sent to the system.
        """


class IButton(IField):
    """A button in a form."""

    access_key = TextLine(title=_('Access Key'),
                          description=_('The key when pressed causes the button to be pressed'),
                          min_length=1,
                          max_length=1,
                          required=False)

    action_factory = Field(title=_('Action Factory'),
                           description=_('The action factory'),
                           required=False,
                           default=None,
                           missing_value=None)


class IImageButton(IButton):
    """An image button in a form."""

    image = TextLine(title=_('Image Path'),
                     description=_('A relative image path to the root of the resources'),
                     required=True)


class IButtons(ISelectionManager):
    """Button manager."""


class IButtonAction(IAction, IWidget, IFieldWidget):
    """Button action."""


class IButtonHandlers(Interface):
    """A collection of handlers for buttons."""

    def add_handler(self, button, handler):
        """Add a new handler for a button."""

    def get_handler(self, button):
        """Get the handler for the button."""

    def copy(self):
        """Copy this object and return the copy."""

    def __add__(self, other):
        """Add another handlers object.

        During the process a copy of the current handlers object should be
        created and the other one is added to the copy. The return value is
        the copy.
        """


class IButtonHandler(Interface):
    """A handler managed by the button handlers."""

    def __call__(self, form, action):  # pylint: disable=signature-differs
        """Execute the handler."""



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.browser.radio module

This module provides radio widgets.
"""

from zope.interface import implementer_only
from zope.schema.interfaces import IBool, ITitledTokenizedTerm
from zope.schema.vocabulary import SimpleTerm

from pyams_form.browser.widget import HTMLInputWidget, add_field_class
from pyams_form.interfaces.widget import IFieldWidget, IRadioWidget
from pyams_form.util import to_unicode
from pyams_form.widget import FieldWidget, SequenceWidget
from pyams_layer.interfaces import IFormLayer
from pyams_template.interfaces import IPageTemplate


__docformat__ = 'restructuredtext'

from pyams_utils.adapter import adapter_config


@implementer_only(IRadioWidget)
class RadioWidget(HTMLInputWidget, SequenceWidget):
    """Input type raiod widget implementation"""

    klass = 'radio-widget'
    css = 'radio'

    def is_checked(self, term):
        """Check if given term is checked"""
        return term.token in self.value

    def render_for_value(self, value):
        """Render given value"""
        terms = list(self.terms)
        try:
            term = self.terms.getTermByToken(value)
        except LookupError:
            if value == SequenceWidget.no_value_token:
                term = SimpleTerm(value)
                terms.insert(0, term)
                item_id = '%s-novalue' % self.id
            else:
                raise
        else:
            item_id = '%s-%i' % (self.id, terms.index(term))
        checked = self.is_checked(term)
        item = {
            'id': item_id,
            'name': self.name,
            'value': term.token,
            'checked': checked
        }
        template = self.request.registry.getMultiAdapter(
            (self.context, self.request, self.form, self.field, self),
            IPageTemplate, name=self.mode + '-single')
        return template(**{
            'context': self.context,
            'request': self.request,
            'view': self,
            'item': item
        })

    @property
    def items(self):
        """Items list getter"""
        if self.terms is None:
            return
        for count, term in enumerate(self.terms):
            checked = self.is_checked(term)
            item_id = '%s-%i' % (self.id, count)
            if ITitledTokenizedTerm.providedBy(term):
                translate = self.request.localizer.translate
                label = translate(term.title)
            else:
                label = to_unicode(term.value)
            yield {
                'id': item_id,
                'name': self.name,
                'value': term.token,
                'label': label,
                'checked': checked
            }

    def update(self):
        """See z3c.form.interfaces.IWidget."""
        super().update()
        add_field_class(self)

    def json_data(self):
        """Get widget data in JSON format"""
        data = super().json_data()
        data['options'] = list(self.items)
        data['type'] = 'radio'
        return data


@adapter_config(required=(IBool, IFormLayer),
                provides=IFieldWidget)
def RadioFieldWidget(field, request):  # pylint: disable=invalid-name
    """IFieldWidget factory for RadioWidget."""
    return FieldWidget(field, RadioWidget(request))



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.form module

This module defines all main forms management classes.
"""

import json
import sys

from persistent import IPersistent
from pyramid.decorator import reify
from pyramid.events import subscriber
from pyramid.response import Response
from zope.interface import implementer
from zope.lifecycleevent import Attributes, ObjectCreatedEvent, ObjectModifiedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty

from pyams_form.button import Buttons, Handlers, button_and_handler
from pyams_form.events import DataExtractedEvent, FormCreatedEvent, FormUpdatedEvent
from pyams_form.field import Fields
from pyams_form.interfaces import DISPLAY_MODE, INPUT_MODE
from pyams_form.interfaces.button import IActionErrorEvent, IActions, WidgetActionExecutionError
from pyams_form.interfaces.error import IErrorViewSnippet
from pyams_form.interfaces.form import IActionForm, IAddForm, IButtonForm, IDisplayForm, \
    IEditForm, IFieldsForm, IForm, IFormAware, IFormContent, IFormFields, IGroup, IGroupManager, \
    IHandlerForm, IInnerSubForm, IInnerTabForm, IInputForm
from pyams_form.interfaces.widget import IWidgets
from pyams_form.util import changed_field
from pyams_security.interfaces.base import FORBIDDEN_PERMISSION
from pyams_security.permission import get_edit_permission
from pyams_template.template import get_content_template, get_layout_template
from pyams_utils.adapter import ContextRequestAdapter
from pyams_utils.factory import create_object
from pyams_utils.interfaces import ICacheKeyValue
from pyams_utils.interfaces.form import IDataManager, NOT_CHANGED
from pyams_utils.url import absolute_url
from pyams_utils.zodb import volatile_property


__docformat__ = 'restructuredtext'

from pyams_form import _  # pylint: disable=ungrouped-imports


REDIRECT_STATUS_CODES = (300, 301, 302, 303, 304, 305, 307)


def get_form_weight(form):
    """Try to get form weight attribute"""
    try:
        return form.weight
    except AttributeError:
        return 0


def apply_changes(form, content, data):
    """Apply form changes to content"""
    data = data.get(form, data)
    changes = {}
    registry = form.request.registry
    for name, field in form.fields.items():
        # If the field is not in the data, then go on to the next one
        try:
            new_value = data[name]
        except KeyError:
            continue
        # If the value is NOT_CHANGED, ignore it, since the widget/converter
        # sent a strong message not to do so.
        if new_value is NOT_CHANGED:
            continue
        if changed_field(field.field, new_value, context=content):
            # Only update the data, if it is different
            dman = registry.getMultiAdapter((content, field.field), IDataManager)
            dman.set(new_value)
            # Record the change using information required later
            changes.setdefault(dman.field.interface, []).append(name)
    return changes


def extends(*args, **kwargs):
    """Extend fields, buttons and handlers

    Please note that extending fields only works if parent class is using static
    fields definition!
    """
    frame = sys._getframe(1)  # pylint: disable=protected-access
    f_locals = frame.f_locals
    if not kwargs.get('ignore_fields', False):
        f_locals['fields'] = Fields()
        for arg in args:
            f_locals['fields'] += getattr(arg, 'fields', Fields())
    if not kwargs.get('ignore_buttons', False):
        f_locals['buttons'] = Buttons()
        for arg in args:
            f_locals['buttons'] += getattr(arg, 'buttons', Buttons())
    if not kwargs.get('ignore_handlers', False):
        f_locals['handlers'] = Handlers()
        for arg in args:
            f_locals['handlers'] += getattr(arg, 'handlers', Handlers())


def merge_changes(source, changes):
    """Merge new changes with existing ones"""
    for intf, field_names in changes.items():
        source[intf] = source.get(intf, []) + field_names


@subscriber(IActionErrorEvent)
def handle_action_error(event):
    """Action error subscriber"""
    # Only react to the event, if the form is a standard form.
    if not (IFormAware.providedBy(event.action) and IForm.providedBy(event.action.form)):
        return
    # If the error was widget-specific, look up the widget.
    widget = None
    if isinstance(event.error, WidgetActionExecutionError):
        widget = event.action.form.widgets[event.error.widget_name]
    # Create an error view for the error.
    action = event.action
    form = action.form
    registry = form.request.registry
    error_view = registry.getMultiAdapter((event.error.error, action.request, widget,
                                           getattr(widget, 'field', None), form,
                                           form.get_content()),
                                          IErrorViewSnippet)
    error_view.update()
    # Assign the error view to all necessary places.
    if widget:
        widget.error = error_view
    form.widgets.errors += (error_view,)
    # If the form supports the ``form_errors_message`` attribute, then set the
    # status to it.
    if hasattr(form, 'form_errors_message'):
        form.status = form.form_errors_message


@implementer(IForm, IFieldsForm)
class BaseForm(ContextRequestAdapter):
    """A base form."""

    title = None
    legend = None
    required_label = _('<span class="required">*</span>&ndash; required')

    prefix = 'form.'
    status = ''
    widgets = None

    template = get_content_template()
    layout = get_layout_template()

    _mode = INPUT_MODE
    _edit_permission = None

    ignore_context = False
    ignore_request = False
    ignore_readonly = False
    ignore_required_on_extract = False

    @volatile_property
    def fields(self):
        """Form fields getter"""
        fields = self.request.registry.queryMultiAdapter((self.context, self.request, self),
                                                         IFormFields)
        if fields is None:
            fields = Fields()
        return fields

    @property
    def mode(self):
        """Current mode getter"""
        mode = self._mode
        if mode == DISPLAY_MODE:  # already custom mode
            return mode
        permission = self.edit_permission
        if permission:
            if permission == FORBIDDEN_PERMISSION:
                mode = DISPLAY_MODE
            else:
                content = self.get_content()
                if not self.request.has_permission(permission, context=content):
                    mode = DISPLAY_MODE
        return mode

    @mode.setter
    def mode(self, value):
        """Form mode setter"""
        self._mode = value

    @volatile_property
    def edit_permission(self):
        """Permission required to access form in input mode"""
        permission = self._edit_permission
        if permission is not None:  # locally defined edit permission
            return permission
        content = self.get_content()
        return get_edit_permission(self.request, content, self)

    @reify
    def form_content(self):
        """Form content getter"""
        content = self.request.registry.queryMultiAdapter((self.context, self.request, self),
                                                          IFormContent)
        if content is None:
            content = self.context
        return content

    def get_content(self):
        """See interfaces.form.IForm"""
        return self.form_content

    @property
    def required_info(self):
        """Form required information label"""
        if (self.required_label is not None and self.widgets is not None and
                self.widgets.has_required_fields):
            return self.request.localizer.translate(self.required_label)
        return None

    @reify
    def subforms(self):
        """Get list of internal sub-forms"""
        registry = self.request.registry
        return sorted((
            adapter
            for name, adapter in registry.getAdapters((self.context, self.request, self),
                                                      IInnerSubForm)
        ), key=get_form_weight)

    @reify
    def tabforms(self):
        """Get list of internal tab-forms"""
        registry = self.request.registry
        return sorted((
            adapter
            for name, adapter in registry.getAdapters((self.context, self.request, self),
                                                      IInnerTabForm)
        ), key=get_form_weight)

    def get_forms(self, include_self=True):
        """Get all forms associated with this form"""
        if include_self:
            yield self
        manager = IGroupManager(self, None)
        if manager is not None:
            for group in manager.groups:
                yield from group.get_forms()
        for form in self.subforms:
            if IInnerSubForm.providedBy(form):
                yield from form.get_forms()
        for form in self.tabforms:
            if IInnerTabForm.providedBy(form):
                yield from form.get_forms()

    def update(self):
        """See interfaces.form.IForm"""
        self.update_widgets()
        [subform.update() for subform in self.subforms]  # pylint: disable=expression-not-assigned
        [tabform.update() for tabform in self.tabforms]  # pylint: disable=expression-not-assigned
        self.request.registry.notify(FormUpdatedEvent(self))

    def update_widgets(self, prefix=None):
        """See interfaces.form.IForm"""
        registry = self.request.registry
        self.widgets = registry.getMultiAdapter((self, self.request, self.get_content()),
                                                IWidgets)
        if prefix is not None:
            self.widgets.prefix = prefix
        self.widgets.mode = self.mode
        self.widgets.ignore_context = self.ignore_context
        self.widgets.ignore_request = self.ignore_request
        self.widgets.ignore_readonly = self.ignore_readonly
        self.widgets.update()

    def extract_data(self, set_errors=True):
        """See interfaces.form.IForm"""
        self.widgets.set_errors = set_errors
        self.widgets.ignore_required_on_extract = self.ignore_required_on_extract
        data, errors = self.widgets.extract()
        self.request.registry.notify(DataExtractedEvent(data, errors, self))
        if not errors:
            errors = self.widgets.errors
        return data, errors

    def get_errors(self):
        """Get all errors, including groups and inner forms"""
        for form in self.get_forms():
            yield from form.widgets.errors

    def render(self):
        """See interfaces.form.IForm"""
        return self.template()

    def __call__(self, **kwargs):
        """Call update and return layout template"""
        self.request.registry.notify(FormCreatedEvent(self))
        self.update()

        # Don't render anything if we are doing a redirect
        request = self.request
        if request.response.status_code in REDIRECT_STATUS_CODES:
            return Response('')

        return Response(self.layout(**kwargs))

    def json(self):
        """Get form data in JSON format"""
        data = {
            'errors': [
                error.message
                for error in (self.widgets.errors or [])
                if error.field is None
            ],
            'prefix': self.prefix,
            'status': self.status,
            'mode': self.mode,
            'fields': [
                widget.json_data()
                for widget in self.widgets.values()
            ],
            'title': self.title or '',
            'legend': self.legend or ''
        }
        return json.dumps(data)


@implementer(IDisplayForm)
class DisplayForm(BaseForm):
    """Display only form"""

    _mode = DISPLAY_MODE
    ignore_request = True


@implementer(IInputForm, IButtonForm, IHandlerForm, IActionForm)
class Form(BaseForm):
    """The Form."""

    buttons = Buttons()

    method = FieldProperty(IInputForm['method'])
    enctype = FieldProperty(IInputForm['enctype'])
    accept_charset = FieldProperty(IInputForm['accept_charset'])
    accept = FieldProperty(IInputForm['accept'])
    autocomplete = FieldProperty(IInputForm['autocomplete'])

    actions = FieldProperty(IActionForm['actions'])
    refresh_actions = FieldProperty(IActionForm['refresh_actions'])

    # AJAX related form properties
    ajax_form_handler = FieldProperty(IInputForm['ajax_form_handler'])
    ajax_form_options = FieldProperty(IInputForm['ajax_form_options'])
    ajax_form_target = FieldProperty(IInputForm['ajax_form_target'])
    ajax_form_callback = FieldProperty(IInputForm['ajax_form_callback'])

    # common string for use in validation status messages
    form_errors_message = _('There were some errors.')

    def __init__(self, context, request):
        super().__init__(context, request)
        self.finished_state = {}

    @property
    def action(self):
        """See interfaces.IInputForm"""
        return self.request.url

    @property
    def name(self):
        """See interfaces.IInputForm"""
        return self.prefix.strip('.')

    @property
    def id(self):  # pylint: disable=invalid-name
        """Form ID"""
        return self.name.replace('.', '-')

    def update_actions(self):
        """Update form actions"""
        registry = self.request.registry
        self.actions = registry.getMultiAdapter((self, self.request, self.get_content()),
                                                IActions)
        self.actions.update()

    def update(self):
        super().update()
        self.update_actions()
        self.actions.execute()
        if self.refresh_actions:
            self.update_actions()

    def get_ajax_handler(self):
        """Get absolute URL of AJAX handler"""
        return absolute_url(self.context, self.request, self.ajax_form_handler)

    def get_form_options(self):
        """Get form options in JSON format"""
        return json.dumps(self.ajax_form_options) if self.ajax_form_options else None


@implementer(IAddForm)
class AddForm(Form):
    """A field and button based add form."""

    ignore_context = True
    ignore_readonly = True

    content_factory = None

    @button_and_handler(_('Add'), name='add')
    def handle_add(self, action):  # pylint: disable=unused-argument
        """Handler for *add* button"""
        data, errors = {}, {}
        for form in self.get_forms():
            form_data, form_errors = form.extract_data()
            if form_errors:
                if not IGroup.providedBy(form):
                    form.status = getattr(form, 'form_errors_message', self.form_errors_message)
                errors[form] = form_errors
            data[form] = form_data
        if errors:
            self.status = self.form_errors_message
            return
        obj = self.create_and_add(data)
        if obj is not None:
            # mark only as finished if we get the new object
            self.finished_state.update({
                'action': action,
                'changes': obj
            })

    def create_and_add(self, data):
        """Create new content and add it to context"""
        obj = self.create(data.get(self, {}))
        self.request.registry.notify(ObjectCreatedEvent(obj))
        if IPersistent.providedBy(obj):  # temporary locate to fix raising of INotYet exceptions
            locate(obj, self.context)
        self.update_content(obj, data)
        self.add(obj)
        return obj

    def create(self, data):  # pylint: disable=unused-argument
        """Create new content from form data"""
        if self.content_factory is not None:
            return create_object(self.content_factory)
        raise NotImplementedError

    def add(self, obj):  # pylint: disable=redefined-builtin
        """Add new object to form context"""
        raise NotImplementedError

    def update_content(self, obj, data):
        """Update content with form data after creation"""
        changes = {}
        for form in self.get_forms():
            if form.mode == DISPLAY_MODE:
                continue
            merge_changes(changes, apply_changes(form, obj, data))
        return changes

    def next_url(self):
        """Redirection URL after object creation"""
        return self.action

    def render(self):
        if self.finished_state:
            self.request.response.location = self.next_url()
            self.request.response.status = 302
            return ''
        return super().render()


@implementer(IEditForm)
class EditForm(Form):
    """A simple edit form with an apply button."""

    success_message = _('Data successfully updated.')
    no_changes_message = _('No changes were applied.')

    @button_and_handler(_('Apply'), name='apply')
    def handle_apply(self, action):  # pylint: disable=unused-argument
        """Apply action handler"""
        data, errors = {}, {}
        for form in self.get_forms():
            form_data, form_errors = form.extract_data()
            if form_errors:
                if not IGroup.providedBy(form):
                    form.status = getattr(form, 'form_errors_message', self.form_errors_message)
                errors[form] = form_errors
            data[form] = form_data
        if errors:
            self.status = self.form_errors_message
            return
        changes = self.apply_changes(data)
        if changes:
            self.status = self.success_message
        else:
            self.status = self.no_changes_message
        self.finished_state.update({
            'action': action,
            'changes': changes
        })

    def apply_changes(self, data):
        """Apply updates to form context"""
        changes = {}
        contents, changed_contents = {}, {}
        for form in self.get_forms():
            if form.mode == DISPLAY_MODE:
                continue
            content = form.get_content()
            form_changes = apply_changes(form, content, data)
            if form_changes:
                merge_changes(changes, form_changes)
                content_hash = ICacheKeyValue(content)
                contents[content_hash] = content
                merge_changes(changed_contents.setdefault(content_hash, {}), form_changes)
        if changes:
            # Construct change-descriptions for the object-modified event
            for content_hash, content_changes in changed_contents.items():
                descriptions = []
                for interface, names in content_changes.items():
                    descriptions.append(Attributes(interface, *names))
                # Send out a detailed object-modified event
                self.request.registry.notify(ObjectModifiedEvent(contents[content_hash],
                                                                 *descriptions))
        return changes


class FormSelector:
    """Form event selector

    This predicate can be used by subscribers to filter form events
    """

    def __init__(self, ifaces, config):  # pylint: disable=unused-argument
        if not isinstance(ifaces, (list, tuple)):
            ifaces = (ifaces,)
        self.interfaces = ifaces

    def text(self):
        """Form's selector text"""
        return 'form_selector = %s' % str(self.interfaces)

    phash = text

    def __call__(self, event):
        for intf in self.interfaces:
            try:
                if intf.providedBy(event.form):
                    return True
            except (AttributeError, TypeError):
                if isinstance(event.form, intf):
                    return True
        return False
